/* queue.c -- FIFO XORed Pointer List
 * 
 * Copyright (C) 2004 H Zero Seven
 *
 * Authors: Christian Volmering <risk@h07.org>
 * 
 * This file is part of the aexpl-ng distribution.
 * 
 * The aexpl-ng distribution is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 of 
 * the License, or (at your option) any later version.
 *
 * The aexpl-ng distribution is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the aexpl-ng distribution; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 * MA 02111-1307 USA
 */

#include "config.h"
#include "queue.h"

#include <stdlib.h>

#ifdef WITH_DMALLOC
#  include <dmalloc.h>
#endif

struct _node_t
{
  const void *ptr;
  unsigned long link;
};

struct _queue_t
{
  unsigned int size;
  unsigned int max_size;
  struct _node_t *base;
  struct _node_t *tail;
  void (*notify) (const void *, int);
};

/* {{{ new_queue */
struct _queue_t *
new_queue (int max_size)
{
  struct _queue_t *queue;

  queue = (struct _queue_t *) malloc (sizeof (struct _queue_t));
  queue->size = 0;
  queue->max_size = max_size;
  queue->base = queue->tail = NULL;
  queue->notify = NULL;

  return (queue);
}

/* }}} */

/* {{{ free_queue */
void
free_queue (struct _queue_t *queue)
{
  clear (queue);
  free (queue);
}

/* }}} */

/* {{{ empty */
int
empty (struct _queue_t *queue)
{
  return (queue->base ? 0 : 1);
}

/* }}} */

/* {{{ size */
int
size (struct _queue_t *queue)
{
  return (queue->size);
}

/* }}} */

/* {{{ clear */
void
clear (struct _queue_t *queue)
{
  int i = size (queue);

  for (; i > 0; --i)
    dequeue (queue);
}

/* }}} */

/* {{{ notify */
void
notify (struct _queue_t *queue, void (*notify) (const void *, int action))
{
  queue->notify = notify;
}

/* }}} */

/* {{{ _notify */
static inline void
_notify (struct _queue_t *queue, const void *ptr, int action)
{
  if (queue->notify)
    (*queue->notify) (ptr, action);
}

/* }}} */

/* {{{ enqueue */
int
enqueue (struct _queue_t *queue, const void *ptr)
{
  struct _node_t *node = (struct _node_t *) malloc (sizeof (struct _node_t));

  if (!node)
    return (0);

  node->ptr = ptr;
  node->link = 0;

  if (!empty (queue))
    {
      node->link ^= (long) queue->tail;
      queue->tail->link ^= (long) node;
      queue->tail = node;
    }
  else
    queue->tail = queue->base = node;

  queue->size++;

  _notify (queue, ptr, NOTIFY_ENQUEUE);

  if (queue->max_size)
    {
      if (queue->size > queue->max_size)
	dequeue (queue);
    }

  return (1);
}

/* }}} */

/* {{{ dequeue */
const void *
dequeue (struct _queue_t *queue)
{
  struct _node_t *node = queue->base;
  const void *ptr = node ? node->ptr : NULL;

  if (node)
    {
      if ((queue->base = (struct _node_t *) (node->link ^ 0)))
	queue->base->link ^= (long) node;
      else
	queue->tail = NULL;

      queue->size--;
      free (node);

      _notify (queue, ptr, NOTIFY_DEQUEUE);
    }

  return (ptr);
}

/* }}} */

/*
 * Local variables:
 * tab-width: 8
 * c-basic-offset: 8
 * End:
 * vim600: noet nonu sw=8 ts=8 fdm=marker
 * vim<600: noet nonu sw=8 ts=8
 */
